
; The  Emulator (C) 1989-91 Todor Todorov
; 6800 and i/o emulating routines

; 28 Sep 1991
; 15 Oct 1991

	.radix	16

trap	=	60		;Trap Interrupt

crtaddr =	0e600		;CRT address register
crtdata =	0e601		;CRT data register
piadra	=	0e628		;PIA data register A
piadrb	=	0e629		;PIA data register B
piacra	=	0e62a		;PIA control register A
piacrb	=	0e62bh		;PIA control register B
jmpadr	=	0e67e		;Call an 8088 routine
rompage =	0e6f0		;Current ROM page

emu	segment
	assume	cs:emu, ds:nothing, es:nothing, ss: stk

; SI = PC
; DI = X
; BP = S
; DH = A
; DL = B
; AH = Carry
; AL = Temp (Opcode)
; CX = Temp (Jump adrs, CL = 0)
; BX = Temp (Eff adrs)

jmpnxt	macro			;Finish the instruction
	lodsb
	mov	ch,al
	jmp	cx
	endm

endnxt	macro			;End the instruction
i	=	i+100
	.errnz	($-emu)/i
	org	i
	endm


next	macro			;Finish + End the instruction
	jmpnxt
	endnxt
	endm
i	=	0

setc	macro			;Clear/Set Carry flag
	local	set_c
	jc	set_c
	mov	ah,cl
	jmpnxt
set_c:
	mov	ah,1
	next
	endm

setcz	macro			;Clear/Set Carry flag (AH = 0)
	local	set_c
	jc	set_c
	jmpnxt
set_c:
	mov	ah,1
	next
	endm

seth	macro			;Clear/Set Half Carry flag
	lahf
	mov	al,ah
	and	al,10
	shl	al,1
	and	byte ptr cs:[flags],0df
	or	cs:[flags],al
	sahf
	endm

sethc	macro			;Clear/Set Carry and Half Carry flags
	local	set_c, set_v, set_c_v
	jo	set_v
	jc	set_c
	seth
	mov	ah,cl
	jmpnxt
set_c:
	seth
	mov	ah,1
	jmpnxt
set_v:
	jc	set_c_v
	seth
	mov	ah,cl
	rcl	ah,1
	jmpnxt
set_c_v:
	seth
	mov	ah,80
	rol	ah,1
	next
	endm

branch	macro			;Take the branch
	cbw
	add	si,ax
	endm

branchf macro			;Take the branch, preserving the flags
	pushf
	mov	ch,ah
	branch
	mov	ah,ch
	popf
	endm

ins	macro			;Increment stack pointer
	pushf
	inc	bp
	popf
	next
	endm

des	macro			;Decrement stack pointer
	pushf
	dec	bp
	popf
	next
	endm

ldfl	macro	reg		;Load flags register
	mov	al,reg
	and	al,30
	or	al,0c0
	mov	cs:[flags],al
	mov	ah,reg
	mov	al,ah
	shl	ax,1
	shl	ax,1
	shl	al,1
	shl	al,1
	and	ax,8c0
	or	ah,2
	push	ax		;---- O-I- SZ-A ---C
	mov	ah,reg
	and	ah,1
	popf
	next
	endm

tstc	macro	reg		;Clear/Set Carry flag, Clear/Set Negative
	local	tst_v		;  and Zero flags according to reg
	jo	tst_v
	rcl	ah,1
	or	reg,reg
	jmpnxt
tst_v:
	rcl	ah,1
	or	reg,reg
	mov	ch,80
	rol	ch,1
	next
	endm

svfl	macro	reg		;Save flags register
	local	get_fl
	pushf
	mov	al,ah
	lahf
	mov	reg,2
	jo	get_fl
	mov	reg,cl
get_fl:
	or	reg,cs:[flags]
	or	reg,al
	xchg	al,ah
	shr	al,1
	shr	al,1
	shr	al,1
	shr	al,1
	and	al,0c
	or	reg,al
	endm

index	macro			;Handle index addressing mode
	lodsb
	mov	bl,al
	mov	bh,cl
	endm

indexz	macro			;Index addressing (AH <- 0)
	lodsb
	mov	ah,cl
	mov	bx,ax
	endm

extend	macro			;Handle extended addressing mode
	lodsw
	mov	bl,ah
	mov	bh,al
	endm

direct	macro			;Handle direct addressing mode
	lodsb
	mov	bl,al
	mov	bh,cl
	endm

directz macro			;Direct addressing (AH <- 0)
	lodsb
	mov	ah,cl
	mov	bx,ax
	endm

chkext	macro	reg		;Check store with extended addressing mode
	local	store_io, store_page, store_vid
	cmp	bh,0e0
	jc	store_page
	cmp	bh,0f0
	jnc	store_vid
	cmp	bh,high rompage
	jz	store_io
	mov	[bx],reg
	or	reg,reg
	jmpnxt
store_io:
	mov	al,reg
	jmp	far ptr write_io
store_page:
	mov	al,reg
	jmp	far ptr write_page
store_vid:
	mov	al,reg
	jmp	far ptr write_vid
	endnxt
	endm

chkind	macro	reg		;Check store with index addressing mode
	local	chk
	add	bx,di
	cmp	bh,0c0
	jnc	chk
	mov	[bx],reg
	or	reg,reg
	jmpnxt
chk:
	chkext	reg
	endm

	next

no_op:				;NOP
	next
	next
	next
	next
	next

	ldfl	dh		;TAP

	svfl	dh		;TPA
	popf
	next

	mov	al,ah		;INX
	lahf
	jo	inc_v
	inc	di
	jz	inc_set_z
	and	ah,0bf
	sahf
	mov	ah,al
	jmpnxt
inc_set_z:
	or	ah,40
	sahf
	mov	ah,al
	jmpnxt
inc_v:
	inc	di
	jz	inc_set_z_v
	and	ah,0bf
	mov	ch,7f
	inc	ch
	sahf
	mov	ah,al
	jmpnxt
inc_set_z_v:
	or	ah,40
	and	ah,0bf
	mov	ch,7f
	inc	ch
	sahf
	mov	ah,al
	next

	mov	al,ah		;DEX
	lahf
	jo	dec_v
	dec	di
	jz	dec_set_z
	and	ah,0bf
	sahf
	mov	ah,al
	jmpnxt
dec_set_z:
	or	ah,40
	sahf
	mov	ah,al
	jmpnxt
dec_v:
	dec	di
	jz	dec_set_z_v
	and	ah,0bf
	mov	ch,7f
	inc	ch
	sahf
	mov	ah,al
	jmpnxt
dec_set_z_v:
	or	ah,40
	and	ah,0bf
	mov	ch,7f
	inc	ch
	sahf
	mov	ah,al
	next

	rol	cl,1		;CLV
	next

	mov	ch,80		;SEV
	rol	ch,1
	next

	mov	ah,cl		;CLC
	next

	mov	ah,1		;SEC
	next

	pushf			;CLI
	and	byte ptr cs:[flags],0ef
	popf
	next

	pushf			;SEI
	or	byte ptr cs:[flags],10
	popf
	next

	sub	dh,dl		;SBA
	setc

	cmp	dh,dl		;CBA
	setc
	next
	next

	and	dh,dl		;NBA -- 6801
	next
	next

	mov	dl,dh		;TAB
	or	dl,dl
	next

	mov	dh,dl		;TBA
	or	dl,dl
	next
	next

	mov	al,cs:[flags]	;DAA
	shr	al,1
	and	al,10
	or	ah,al
	sahf
	mov	al,dh
	daa
	mov	dh,al
	setc

	or	dh,dl		;OBA -- 6801
	next

	add	dh,dl		;ABA
	sethc
	next
	next
	next
	next

	lodsb			;BRA -
	branchf
	next
	next

	lodsb			;BHI -
	jz	no_bhi
	pushf
	test	ah,ah
	jnz	no_bhi_c
	branch
	mov	ah,cl
no_bhi_c:
	popf
no_bhi:
	next

	lodsb			;BLS -
	pushf
	jz	bls_do
	test	ah,ah
	jz	no_bls
bls_do:
	branch
	mov	ah,1
no_bls:
	popf
	next

	lodsb			;BCC -
	pushf
	test	ah,ah
	jnz	no_bcc
	branch
	mov	ah,cl
no_bcc:
	popf
	next

	lodsb			;BCS -
	pushf
	test	ah,ah
	jz	no_bcs
	branch
	mov	ah,1
no_bcs:
	popf
	next

	lodsb			;BNE -
	jz	no_bne
	branchf
no_bne:
	next

	lodsb			;BEQ -
	jnz	no_beq
	branchf
no_beq:
	next

	lodsb			;BVC -
	jo	no_bvc
	branchf
no_bvc:
	next

	lodsb			;BVS -
	jno	no_bvs
	branchf
no_bvs:
	next

	lodsb			;BPL -
	js	no_bpl
	branchf
no_bpl:
	next

	lodsb			;BMI -
	jns	no_bmi
	branchf
no_bmi:
	next

	lodsb			;BGE -
	jl	no_bge
	branchf
no_bge:
	next

	lodsb			;BLT -
	jge	no_blt
	branchf
no_blt:
	next

	lodsb			;BGT -
	jle	no_bgt
	branchf
no_bgt:
	next

	lodsb			;BLE -
	jg	no_ble
	branchf
no_ble:
	next

	lea	di,[bp+1]	;TSX
	next

	ins			;INS

	mov	dh,ds:[bp+1]	;PULA
	ins

	mov	dl,ds:[bp+1]	;PULB
	ins

	des			;DES

	lea	bp,[di-1]	;TXS
	next

	mov	ds:[bp],dh	;PSHA
	des

	mov	ds:[bp],dl	;PSHB
	des
	next

	mov	bx,ds:[bp+1]	;RTS
	xchg	bl,bh
	mov	si,bx
	pushf
	inc	bp
	inc	bp
	popf
	next
	next

	add	bp,7		;RTI
	mov	bx,ds:[bp-1]
	xchg	bl,bh
	mov	si,bx
	mov	bx,ds:[bp-3]
	xchg	bl,bh
	mov	di,bx
	mov	dx,ds:[bp-5]
	ldfl	ds:[bp-6]
	next
	next

	int	trap		;WAI -- Breakpoint
	next			;????

swi:
	mov	bx,si		;SWI
	xchg	bl,bh
	mov	ds:[bp-1],bx
	mov	bx,di
	xchg	bl,bh
	mov	ds:[bp-3],bx
	mov	ds:[bp-5],dx
	svfl	ch
	mov	ds:[bp-6],ch
	sub	bp,7
	or	byte ptr cs:[flags],10
	popf
	mov	bx,ds:[-6]
	xchg	bl,bh
	mov	si,bx
	next

	neg	dh		;NEGA
	setc
	next
	next

	not	dh		;COMA
	or	dh,dh
	mov	ah,1
	next

	shr	dh,1		;LSRA
	setc
	next

	shr	ah,1		;RORA
	rcr	dh,1
	tstc	dh

	sar	dh,1		;ASRA
	setc

	sal	dh,1		;ASLA
	setc

	shr	ah,1		;ROLA
	rcl	dh,1
	tstc	dh

	dec	dh		;DECA
	next
	next

	inc	dh		;INCA
	next

	or	dh,dh		;TSTA
	mov	ah,cl
	next
	next

	xor	dh,dh		;CLRA
	mov	ah,cl
	next

	neg	dl		;NEGB
	setc
	next
	next

	not	dl		;COMB
	or	dl,dl
	mov	ah,1
	next

	shr	dl,1		;LSRB
	setc
	next

	shr	ah,1		;RORB
	rcr	dl,1
	tstc	dl

	sar	dl,1		;ASRB
	setc

	sal	dl,1		;ASLB
	setc

	shr	ah,1		;ROLB
	rcl	dl,1
	tstc	dl

	dec	dl		;DECB
	next
	next

	inc	dl		;INCB
	next

	or	dl,dl		;TSTB
	mov	ah,cl
	next
	next

	xor	dl,dl		;CLRB
	mov	ah,cl
	next

	indexz			;NEG -,X
	neg	byte ptr [bx+di]
	setcz
	next
	next

	indexz			;COM -,X
	not	byte ptr [bx+di]
	cmp	[bx+di],cl
	mov	ah,1
	next

	indexz			;LSR -,X
	shr	byte ptr [bx+di],1
	setcz
	next

	index			;ROR -,X
	shr	ah,1
	rcr	byte ptr [bx+di],1
	mov	al,[bx+di]
	tstc	al

	indexz			;ASR -,X
	sar	byte ptr [bx+di],1
	setcz

	indexz			;ASL -,X
	sal	byte ptr [bx+di],1
	setcz

	index			;ROL -,X
	shr	ah,1
	rcl	byte ptr [bx+di],1
	mov	al,[bx+di]
	tstc	al

	index			;DEC -,X
	dec	byte ptr [bx+di]
	next
	next

	index			;INC -,X
	inc	byte ptr [bx+di]
	next

	indexz			;TST -,X
	cmp	[bx+di],cl
	next

	index			;JMP -,X
	lea	si,[bx+di]
	next

	indexz			;CLR -,X
	cmp	di,0bf00
	jnc	clr_chk_i
	mov	[bx+di],cl
	or	cl,cl
	jmpnxt
clr_chk_i:
	chkind	cl

	extend			;NEG --
	neg	byte ptr [bx]
	setc
	next
	next

	extend			;COM --
	not	byte ptr [bx]
	cmp	[bx],cl
	mov	ah,1
	next

	extend			;LSR --
	shr	byte ptr [bx],1
	setc
	next

	shr	ah,1		;ROR --
	extend
	mov	ah,cl
	rcr	byte ptr [bx],1
	mov	al,[bx]
	tstc	al

	extend			;ASR --
	sar	byte ptr [bx],1
	setc

	extend			;ASL --
	sal	byte ptr [bx],1
	setc

	shr	ah,1		;ROL --
	extend
	mov	ah,cl
	rcl	byte ptr [bx],1
	mov	al,[bx]
	tstc	al

	mov	ch,ah		;DEC --
	extend
	mov	ah,ch
	dec	byte ptr [bx]
	next
	next

	mov	ch,ah		;INC --
	extend
	mov	ah,ch
	inc	byte ptr [bx]
	next

	extend			;TST --
	xor	ah,ah
	cmp	[bx],cl
	next

	mov	ch,ah		;JMP --
	extend
	mov	ah,ch
	mov	si,bx
	next

	extend			;CLR --
	mov	ah,cl
	cmp	bh,0c0
	jnc	clr_chk_x
	mov	[bx],cl
	or	cl,cl
	jmpnxt
clr_chk_x:
	chkext	cl

	lodsb			;SUBA #-
	sub	dh,al
	setc

	lodsb			;CMPA #-
	cmp	dh,al
	setc

	lodsb			;SBCA #-
	shr	ah,1
	sbb	dh,al
	setc
	next

	lodsb			;ANDA #-
	and	dh,al
	next

	lodsb			;BITA #-
	test	dh,al
	next

	lodsb			;LDAA #-
	mov	dh,al
	or	dh,dh
	next
	next

	lodsb			;EORA #-
	xor	dh,al
	next

	lodsb			;ADCA #-
	shr	ah,1
	adc	dh,al
	sethc

	lodsb			;ORAA #-
	or	dh,al
	next

	lodsb			;ADDA #-
	add	dh,al
	sethc

	mov	ch,ah		;CPX #--
	lodsw
	xchg	al,ah
	cmp	di,ax
	mov	ah,ch
	next

	lodsb			;BSR -
	mov	bx,si
	xchg	bl,bh
	mov	ds:[bp-1],bx
	pushf
	mov	ch,ah
	dec	bp
	dec	bp
	branch
	mov	ah,ch
	popf
	next

	mov	ch,ah		;LDS #--
	lodsw
	xchg	al,ah
	mov	bp,ax
	mov	ah,ch
	or	bp,bp
	next
	next

	directz 		;SUBA -
	sub	dh,[bx]
	setcz

	directz 		;CMPA -
	cmp	dh,[bx]
	setcz

	direct			;SBCA -
	shr	ah,1
	sbb	dh,[bx]
	setcz
	next

	direct			;ANDA -
	and	dh,[bx]
	next

	direct			;BITA -
	test	dh,[bx]
	next

	direct			;LDAA -
	mov	dh,[bx]
	or	dh,dh
	next

	direct			;STAA -
	mov	[bx],dh
	or	dh,dh
	next

	direct			;EORA -
	xor	dh,[bx]
	next

	direct			;ADCA -
	shr	ah,1
	adc	dh,[bx]
	sethc

	direct			;ORAA -
	or	dh,[bx]
	next

	direct			;ADDA -
	add	dh,[bx]
	sethc

	direct			;CPX -
	mov	al,[bx]
	inc	bl
	mov	bl,[bx]
	mov	bh,al
	cmp	di,bx
	next
	next

	direct			;LDS -
	mov	al,[bx]
	inc	bl
	mov	bl,[bx]
	mov	bh,al
	mov	bp,bx
	or	bp,bp
	next

	direct			;STS -
	mov	cx,bp
	mov	[bx],ch
	inc	bl
	mov	[bx],cl
	xor	cl,cl
	or	bp,bp
	next

	indexz			;SUBA -,X
	sub	dh,[bx+di]
	setcz

	indexz			;CMPA -,X
	cmp	dh,[bx+di]
	setcz

	index			;SBCA -,X
	shr	ah,1
	sbb	dh,[bx+di]
	setcz
	next

	index			;ANDA -,X
	and	dh,[bx+di]
	next

	index			;BITA -,X
	test	dh,[bx+di]
	next

	index			;LDAA -,X
	mov	dh,[bx+di]
	or	dh,dh
	next

	index			;STAA -,X
	cmp	di,0bf00
	jnc	sta_a_chk_i
	mov	[bx+di],dh
	or	dh,dh
	jmpnxt
sta_a_chk_i:
	chkind	dh

	index			;EORA -,X
	xor	dh,[bx+di]
	next

	index			;ADCA -,X
	shr	ah,1
	adc	dh,[bx+di]
	sethc

	index			;ORAA -,X
	or	dh,[bx+di]
	next

	index			;ADDA -,X
	add	dh,[bx+di]
	sethc

	index			;CPX -,X
	mov	bx,[bx+di]
	xchg	bl,bh
	cmp	di,bx
	next

	lea	bx,[si+1]	;JSR -,X
	xchg	bl,bh
	mov	ds:[bp-1],bx
	index
	lea	si,[bx+di]
	pushf
	dec	bp
	dec	bp
	popf
	next

	index			;LDS -,X
	mov	bx,[bx+di]
	xchg	bl,bh
	mov	bp,bx
	or	bp,bp
	next

	index			;STS -,X
	mov	cx,bp
	xchg	cl,ch
	mov	[bx+di],cx
	xor	cl,cl
	or	bp,bp
	next

	extend			;SUBA --
	sub	dh,[bx]
	setc

	extend			;CMPA --
	cmp	dh,[bx]
	setc

	shr	ah,1		;SBCA --
	extend
	sbb	dh,[bx]
	setc
	next

	mov	ch,ah		;ANDA --
	extend
	mov	ah,ch
	and	dh,[bx]
	next

	mov	ch,ah		;BITA --
	extend
	mov	ah,ch
	test	dh,[bx]
	next

	mov	ch,ah		;LDAA --
	extend
	mov	ah,ch
	mov	dh,[bx]
	or	dh,dh
	next

	mov	ch,ah		;STAA --
	extend
	mov	ah,ch
	cmp	bh,0c0
	jnc	sta_a_chk_x
	mov	[bx],dh
	or	dh,dh
	jmpnxt
sta_a_chk_x:
	chkext	dh

	mov	ch,ah		;EORA --
	extend
	mov	ah,ch
	xor	dh,[bx]
	next

	shr	ah,1		;ADCA --
	extend
	adc	dh,[bx]
	sethc

	mov	ch,ah		;ORAA --
	extend
	mov	ah,ch
	or	dh,[bx]
	next

	extend			;ADDA --
	add	dh,[bx]
	sethc

	mov	ch,ah		;CPX --
	extend
	mov	ah,ch
	mov	bx,[bx]
	xchg	bl,bh
	cmp	di,bx
	next

	mov	ch,ah		;JSR --
	extend
	mov	ax,si
	xchg	al,ah
	mov	ds:[bp-1],ax
	mov	ah,ch
	mov	si,bx
	pushf
	dec	bp
	dec	bp
	popf
	next

	mov	ch,ah		;LDS --
	extend
	mov	ah,ch
	mov	bx,[bx]
	xchg	bl,bh
	mov	bp,bx
	or	bp,bp
	next

	mov	ch,ah		;STS --
	extend
	mov	ax,bp
	xchg	al,ah
	mov	[bx],ax
	mov	ah,ch
	or	bp,bp
	next

	lodsb			;SUBB #-
	sub	dl,al
	setc

	lodsb			;CMPB #-
	cmp	dl,al
	setc

	lodsb			;SBCB #-
	shr	ah,1
	sbb	dl,al
	setc
	next

	lodsb			;ANDB #-
	and	dl,al
	next

	lodsb			;BITB #-
	test	dl,al
	next

	lodsb			;LDAB #-
	mov	dl,al
	or	dl,dl
	next
	next

	lodsb			;EORB #-
	xor	dl,al
	next

	lodsb			;ADCB #-
	shr	ah,1
	adc	dl,al
	sethc

	lodsb			;ORAB #-
	or	dl,al
	next

	lodsb			;ADDB #-
	add	dl,al
	sethc
	next
	next

	mov	ch,ah		;LDX #--
	lodsw
	xchg	al,ah
	mov	di,ax
	mov	ah,ch
	or	di,di
	next
	next

	directz 		;SUBB -
	sub	dl,[bx]
	setcz

	directz 		;CMPB -
	cmp	dl,[bx]
	setcz

	direct			;SBCB -
	shr	ah,1
	sbb	dl,[bx]
	setcz
	next

	direct			;ANDB -
	and	dl,[bx]
	next

	direct			;BITB -
	test	dl,[bx]
	next

	direct			;LDAB -
	mov	dl,[bx]
	or	dl,dl
	next

	direct			;STAB -
	mov	[bx],dl
	or	dl,dl
	next

	direct			;EORB -
	xor	dl,[bx]
	next

	direct			;ADCB -
	shr	ah,1
	adc	dl,[bx]
	sethc

	direct			;ORAB -
	or	dl,[bx]
	next

	direct			;ADDB -
	add	dl,[bx]
	sethc
	next
	next

	direct			;LDX -
	mov	al,[bx]
	inc	bl
	mov	bl,[bx]
	mov	bh,al
	mov	di,bx
	or	di,di
	next

	direct			;STX -
	mov	cx,di
	mov	[bx],ch
	inc	bl
	mov	[bx],cl
	xor	cl,cl
	or	di,di
	next

	indexz			;SUBB -,X
	sub	dl,[bx+di]
	setcz

	indexz			;CMPB -,X
	cmp	dl,[bx+di]
	setcz

	index			;SBCB -,X
	shr	ah,1
	sbb	dl,[bx+di]
	setcz
	next

	index			;ANDB -,X
	and	dl,[bx+di]
	next

	index			;BITB -,X
	test	dl,[bx+di]
	next

	index			;LDAB -,X
	mov	dl,[bx+di]
	or	dl,dl
	next

	index			;STAB -,X
	cmp	di,0bf00
	jnc	sta_b_chk_i
	mov	[bx+di],dl
	or	dl,dl
	jmpnxt
sta_b_chk_i:
	chkind	dl

	index			;EORB -,X
	xor	dl,[bx+di]
	next

	index			;ADCB -,X
	shr	ah,1
	adc	dl,[bx+di]
	sethc

	index			;ORAB -,X
	or	dl,[bx+di]
	next

	index			;ADDB -,X
	add	dl,[bx+di]
	sethc
	next
	next

	index			;LDX -,X
	mov	bx,[bx+di]
	xchg	bl,bh
	mov	di,bx
	or	di,di
	next

	index			;STX -,X
	mov	cx,di
	xchg	cl,ch
	mov	[bx+di],cx
	xor	cl,cl
	or	di,di
	next

	extend			;SUBB --
	sub	dl,[bx]
	setc

	extend			;CMPB --
	cmp	dl,[bx]
	setc

	shr	ah,1		;SBCB --
	extend
	sbb	dl,[bx]
	setc
	next

	mov	ch,ah		;ANDB --
	extend
	mov	ah,ch
	and	dl,[bx]
	next

	mov	ch,ah		;BITB --
	extend
	mov	ah,ch
	test	dl,[bx]
	next

	mov	ch,ah		;LDAB --
	extend
	mov	ah,ch
	mov	dl,[bx]
	or	dl,dl
	next

	mov	ch,ah		;STAB --
	extend
	mov	ah,ch
	cmp	bh,0c0
	jnc	sta_b_chk_x
	mov	[bx],dl
	or	dl,dl
	jmpnxt
sta_b_chk_x:
	chkext	dl

	mov	ch,ah		;EORB --
	extend
	mov	ah,ch
	xor	dl,[bx]
	next

	shr	ah,1		;ADCB --
	extend
	adc	dl,[bx]
	sethc

	mov	ch,ah		;ORAB --
	extend
	mov	ah,ch
	or	dl,[bx]
	next

	extend			;ADDB --
	add	dl,[bx]
	sethc
	next
	next

	mov	ch,ah		;LDX --
	extend
	mov	ah,ch
	mov	bx,[bx]
	xchg	bl,bh
	mov	di,bx
	or	di,di
	next

	mov	ch,ah		;STX --
	extend
	cmp	bx,jmpadr
	jz	call_intel
	mov	ax,di
	xchg	al,ah
	mov	[bx],ax
	mov	ah,ch
resume_stx:
	or	di,di
	jmpnxt
call_intel:			;Call an 8088 subroutine
	mov	bx,offset intel_ret
	push	es ds bp di si dx
	pushf
	push	cs bx
	push	ds di
	retf
intel_ret:
	pop	dx si di bp ds es
	mov	ah,1
	xor	cl,cl
	jmp	resume_stx
	jmp	cx		;Must be present for the Debugger!

	org	-1
flags	db	0d0		;Flags register (11HInzvc)

	ends

main	segment
	assume	cs:main

write_vid:			;Write byte to video memory (Fxxx)  AL -> BX
	add	bh,10
	shl	bx,1
	mov	es:[bx],al
	jmp	skip_write

write_page:			;Write byte to page memory (Cxxx-Dxxx)	AL -> BX
	push	ds
	test	byte ptr ds:[rompage],8
	jz	resume_cpu_go
	push	dx
	mov	dx,seg ram-0c00
	mov	ds,dx
	pop	dx
resume_cpu_go:
	mov	[bx],al
	pop	ds
	jmp	skip_write

chk_crt:
	push	dx ax
	mov	ch,ds:[crtaddr]
	sub	ch,0a
	cmp	ch,6
	jnc	skip_write_go
	cmp	ch,4
	jz	xlat_adr
	cmp	ch,2
	jnz	no_adr
xlat_adr:
	add	al,10
no_adr:
	push	es
	mov	es,cs:[crtseg]
	mov	bl,ch
	xor	bh,bh
	mov	es:[bx],al
	pop	es
	mov	bx,es
	cmp	bh,0b8
	jnz	skip_write_go
	cmp	ch,2
	jc	bios_crsr
	xchg	al,ch
	add	al,0a
	mov	dx,3d4
	out	dx,al
	inc	dx
	mov	al,ch
	out	dx,al
	jmp	short skip_write_go
bios_crsr:
	push	bp di si
	push	cx ax
	xor	bh,bh
	mov	ah,3
	int	10
	pop	ax bx
	test	bh,bh
	jz	crsr_start
	mov	cl,al
	jmp	short crsr_end
crsr_start:
	mov	ch,al
crsr_end:
	mov	ah,1
	int	10
	xor	cl,cl
	pop	si di bp
skip_write_go:
	pop	ax dx
	jmp	skip_write

chk_spkr:
	push	ax
	and	al,8
	xchg	al,[bx]
	cmp	al,[bx]
	jz	no_click
	in	al,[61]
	xor	al,2
	out	[61],al
no_click:
	pop	ax
	jmp	skip_write

write_io:			;Write byte to i/o memory (E6xx)  AL -> BX
	cmp	bl,low crtaddr+20
	jnc	not_crt
	and	bl,low crtdata
	jz	resume_cpu
	jmp	chk_crt
not_crt:
	cmp	bl,low piacrb
	jz	chk_spkr
	cmp	bl,low rompage
	jnz	skip_write

move_rom:
	push	es di si dx bx ax
	and	al,0f
	mov	ch,[bx]
	and	ch,0f
	cmp	al,ch
	jz	dont_move
	mov	dx,seg ram
	test	ch,8
	jnz	dont_save
	mov	si,0c000
	mov	es,dx
	xor	di,di
	mov	cx,1000
	rep	movsw
dont_save:
	xor	si,si
	test	al,8
	jz	move_ram
	mov	dx,seg rom
	mov	cl,0dh
	shl	ax,cl
	mov	si,ax
move_ram:
	push	ds
	pop	es
	mov	di,0c000
	mov	ds,dx
	mov	cx,1000
	rep	movsw
	push	es
	pop	ds
dont_move:
	pop	ax bx dx si di es

resume_cpu:
	mov	[bx],al
skip_write:
	or	al,al
	jmp	short jmp_far

start:
	cld
	mov	ax,seg emu
	mov	bx,seg pyldin
	mov	cx,seg ram
	mov	dx,seg rom
	int	trap
	add	ah,2
	mov	cs:[crtseg],ax
	mov	ds,bx
	mov	ax,0b800
	mov	es,ax
	mov	ax,ds:[-2]
	xchg	al,ah
	xchg	ax,si
	mov	bp,0c000
	xor	di,di
	xor	ax,ax
	xor	cx,cx
	xor	dx,dx
jmp_far:
	jmp	far ptr no_op

crtseg	dw	?

	ends

rom	segment
	db	0ffff dup(?)
	ends

pyldin	segment
	db	0ffff dup(?)
	ends

ram	segment
	db	2000 dup(?)
	ends

stk	segment stack 'STACK'
	db	400 dup(?)
	ends

	end	start
